// 一个更加通用的冒泡排序函数的代码和标注
//  要点
//    1 函数 bubbleSortAny（a, m, off) 可以对数组 a 中从 a+off 开始的 m 个元素排序 
//    2 数组一般性质的描述以及与原数据区的关系
//    3 采用量化断言描述数组的有序性

#define MAX_LEN 10000

// 函数协议
//     前条件：数组a 的一般性质：数组（区）的长度 >= m+off，偏移是off。
//     assigns子句：提示数组段 a[0..m-1]的元素（相对于原数据区为[off..m-1+off]）在函数中可能被修改
//     后条件：排序后数组 a[0..m-1] 的元素按递增顺序排列
/*@ 
    requires 
        m+off <= \length(a) <= MAX_LEN && \offset(a) == off && off >= 0 && m > 0
        ;
    assigns 
        a[0..m-1]
        ;
    ensures  
        m+off <= \length(a) <= MAX_LEN && \offset(a) == off && off >= 0 && m > 0
        && (\forall int i : [0..m-2]. a[i] <= a[i+1])
        ;
*/
void bubbleSortAny(int *const a, int const m, int const off) {
    int low, up, j, k, temp;
    low = 0; 
    up = m-1; 
    k = up; 
    //    循环不变式
    //        每次排序都会把数组段[low..k]最大的值放在该数组段的最后一位
    //    循环入口（k == up）
    //        尚未排序
    //    循环中 && 循环出口
    //        每次循环后，会选出当前最大元素a[k+1]，将数组分成2部分：
    //        未排序的数组段[low..k]和已排序的数组段[k+1..up-1]。其中
    //        a[k+1]大于[low..k]中的所有元素，数组段[k+1..up-1]有序
    /*@ loop invariant 
        (k == up
         ||
         low <= k < up && (\forall int i:[low..k].a[i]<=a[k+1]) && \forall int i:[k+1..up-1].a[i]<=a[i+1]
        )
                         // 来自函数前条件的性质
        && low == 0 && up == m-1 && 
        m+off <= \length(a) <= MAX_LEN && \offset(a) == off && off >= 0 && m>0
        ;
    */
    while(k != low) {
        j = low;
        //    循环不变式
        //        在数组段[low..k]内检索，通过交换相邻元素选出最大元素
        //        在一轮比较交换过程中，a[j]总是大于数组段a[low..j-1]内的元素
        //    循环入口（k == up && j == low）
        //    循环中 && 循环出口
        //        同时具有外层循环的性质
        /*@ loop invariant 
            k != low && low <= j <= k && low == 0 && up == m-1
            && (\forall int i:[low..j-1]. a[i]<=a[j]) 
            && // 外层循环的性质
            ( k == up
             || 
             low < k < up && (\forall int i:[low..k].a[i]<=a[k+1]) && (\forall int i:[k+1..up-1].a[i]<=a[i+1])
            )
            // 函数前条件
            && m+off <= \length(a) <= MAX_LEN && \offset(a) == off && off >= 0 && m > 0 
            ;
        */
        while( j!=k ) {
            if (a[j] > a[j+1]){
                temp = a[j]; 
                a[j] = a[j+1]; 
                a[j+1] = temp;
            }
            j = j + 1;
        }
        k = k -1;
    }
}

int main(){
    int arr[] = {3, 2, 1, 4, 6, 7, 5, 8, 10, 9, 20, 18, 19, 16, 15, 13, 14, 17, 12, 11};

    bubbleSortAny(arr+3, 15, 3);  // 排序arr[3]～arr[17]共15个元素
    /*@assert
        arr[0]==3 && arr[1]==2 && arr[2]==1           // 第一段，未排序
        && (\forall int i:[3..16].arr[i]<=arr[i+1])   // 第二段，已排序
        && arr[18]==12 && arr[19]==11                 // 第三段，未排序
        ;
    */
   return 0;
}